<?php
namespace WSCMIME;

/**
 * Class for parsing MIME messages
 *
 * @package    WSCOMUN
 * @author     David PAscual. Based on Aleksander Machniak <alec@alec.pl>
 */

require_once 'WSCMimeParser.php';
require_once 'WSCCharset.php';
require_once 'WSCMessagePart.php';

class WSCMime
{
	const WSC_CHARSET = 'UTF-8';
	private static $default_charset;


	/**
	 * Constructor
	 * @param string $default_charset
	 */
	function __construct($default_charset = null)
	{
		self::$default_charset = self::WSC_CHARSET;
		if (!empty($default_charset))
		{
			self::$default_charset = $default_charset;
		}
	}


	/**
	 * Returns message/object character set name
	 *
	 * @return string Characted set name
	 */
	public static function get_charset()
	{
		if (self::$default_charset)
		{
			return self::$default_charset;
		}
		return self::WSC_CHARSET;
	}


	/**
	 * Parse the given raw message source and return a structure
	 * of message_part.
	 *
	 * It makes use of the WSCMimeParser library
	 *
	 * @param string $raw_body The message source
	 *
	 * @return object WSCMessagePart The message structure
	 */
	public static function parse_message($raw_body)
	{
		$conf = array(
				'include_bodies'  => true,
				'decode_bodies'   => true,
				'decode_headers'  => false,
				'default_charset' => self::get_charset(),
		);
		$mime = new WSCMimeParser($conf);
		return $mime->decode($raw_body);
	}

	/**
	 * Split an address list into a structured array list
	 *
	 * @param string  $input    Input string
	 * @param int     $max      List only this number of addresses
	 * @param boolean $decode   Decode address strings
	 * @param string  $fallback Fallback charset if none specified
	 * @param boolean $addronly Return flat array with e-mail addresses only
	 *
	 * @return array Indexed list of addresses
	 */
	static function decode_address_list($input, $max = null, $decode = true, $fallback = null, $addronly = false)
	{
		$a   = self::parse_address_list($input, $decode, $fallback);
		$out = array();
		$j   = 0;
		// Special chars as defined by RFC 822 need to in quoted string (or escaped).
		$special_chars = '[\(\)\<\>\\\.\[\]@,;:"]';
		if (!is_array($a)) {
			return $out;
		}
		foreach ($a as $val) {
			$j++;
			$address = trim($val['address']);
			if ($addronly) {
				$out[$j] = $address;
			}
			else {
				$name = trim($val['name']);
				if ($name && $address && $name != $address)
					$string = sprintf('%s <%s>', preg_match("/$special_chars/", $name) ? '"'.addcslashes($name, '"').'"' : $name, $address);
					else if ($address)
						$string = $address;
						else if ($name)
							$string = $name;
							$out[$j] = array('name' => $name, 'mailto' => $address, 'string' => $string);
			}
			if ($max && $j==$max)
				break;
		}
		return $out;
	}

	/**
	 * Decode a message header value
	 *
	 * @param string  $input    Header value
	 * @param string  $fallback Fallback charset if none specified
	 *
	 * @return string Decoded string
	 */
	public static function decode_header($input, $fallback = null)
	{
		$str = self::decode_mime_string((string)$input, $fallback);
		return $str;
	}
	/**
	 * Decode a mime-encoded string to internal charset
	 *
	 * @param string $input    Header value
	 * @param string $fallback Fallback charset if none specified
	 *
	 * @return string Decoded string
	 */

	public static function decode_mime_string($input, $fallback = null)
	{
		$matches = array();

		$default_charset = $fallback ?: self::get_charset();
		// rfc: all line breaks or other characters not found
		// in the Base64 Alphabet must be ignored by decoding software
		// delete all blanks between MIME-lines, differently we can
		// receive unnecessary blanks and broken utf-8 symbols
		$input = preg_replace("/\?=\s+=\?/", '?==?', $input);
		// encoded-word regexp
		$re = '/=\?([^?]+)\?([BbQq])\?([^\n]*?)\?=/';
		// Find all RFC2047's encoded words
		if (preg_match_all($re, $input, $matches, PREG_OFFSET_CAPTURE | PREG_SET_ORDER))
		{
			// Initialize variables
			$tmp   = array();
			$out   = '';
			$start = 0;
			foreach ($matches as $idx => $m) {
				$pos      = $m[0][1];
				$charset  = $m[1][0];
				$encoding = $m[2][0];
				$text     = $m[3][0];
				$length   = strlen($m[0][0]);
				// Append everything that is before the text to be decoded
				if ($start != $pos) {
					$substr = substr($input, $start, $pos-$start);
					$out   .= WSCCharset::convert($substr, $default_charset);
					$start  = $pos;
				}
				$start += $length;
				// Per RFC2047, each string part "MUST represent an integral number
				// of characters . A multi-octet character may not be split across
				// adjacent encoded-words." However, some mailers break this, so we
				// try to handle characters spanned across parts anyway by iterating
				// through and aggregating sequential encoded parts with the same
				// character set and encoding, then perform the decoding on the
				// aggregation as a whole.
				$tmp[] = $text;
				if ($next_match = $matches[$idx+1]) {
					if ($next_match[0][1] == $start
							&& $next_match[1][0] == $charset
							&& $next_match[2][0] == $encoding
							) {
								continue;
							}
				}
				$count = count($tmp);
				$text  = '';
				// Decode and join encoded-word's chunks
				if ($encoding == 'B' || $encoding == 'b') {
					// base64 must be decoded a segment at a time
					for ($i=0; $i<$count; $i++)
						$text .= base64_decode($tmp[$i]);
				}
				else { //if ($encoding == 'Q' || $encoding == 'q') {
					// quoted printable can be combined and processed at once
					for ($i=0; $i<$count; $i++)
						$text .= $tmp[$i];
						$text = str_replace('_', ' ', $text);
						$text = quoted_printable_decode($text);
				}
				$out .= WSCCharset::convert($text, $charset);
				$tmp = array();
			}
			// add the last part of the input string
			if ($start != strlen($input)) {
				$out .= WSCCharset::convert(substr($input, $start), $default_charset);
			}
			// return the results
			return $out;
		}
		// no encoding information, use fallback
		return WSCCharset::convert($input, $default_charset);
	}

	/**
	 * Decode a mime part
	 *
	 * @param string $input    Input string
	 * @param string $encoding Part encoding
	 *
	 * @return string Decoded string
	 */
	public static function decode($input, $encoding = '7bit')
	{
		switch (strtolower($encoding)) {
			case 'quoted-printable':
				return quoted_printable_decode($input);
			case 'base64':
				return base64_decode($input);
			case 'x-uuencode':
			case 'x-uue':
			case 'uue':
			case 'uuencode':
				return convert_uudecode($input);
			case '7bit':
			default:
				return $input;
		}
	}

	/**
	 * Split RFC822 header string into an associative array
	 */
	public static function parse_headers($headers)
	{
		$a_headers = array();
		$headers   = preg_replace('/\r?\n(\t| )+/', ' ', $headers);
		$lines     = explode("\n", $headers);
		$count     = count($lines);
		for ($i=0; $i<$count; $i++) {
			if ($p = strpos($lines[$i], ': ')) {
				$field = strtolower(substr($lines[$i], 0, $p));
				$value = trim(substr($lines[$i], $p+1));
				if (!empty($value)) {
					$a_headers[$field] = $value;
				}
			}
		}
		return $a_headers;
	}

	/**
	 * E-mail address list parser
	 */
	private static function parse_address_list($str, $decode = true, $fallback = null)
	{
		$m = array();
		// remove any newlines and carriage returns before
		$str = preg_replace('/\r?\n(\s|\t)?/', ' ', $str);
		// extract list items, remove comments
		$str = self::explode_header_string(',;', $str, true);
		$result = array();
		// simplified regexp, supporting quoted local part
		$email_rx = '(\S+|("\s*(?:[^"\f\n\r\t\v\b\s]+\s*)+"))@\S+';
		foreach ($str as $key => $val) {
			$name    = '';
			$address = '';
			$val     = trim($val);
			if (preg_match('/(.*)<('.$email_rx.')>$/', $val, $m)) {
				$address = $m[2];
				$name    = trim($m[1]);
			}
			else if (preg_match('/^('.$email_rx.')$/', $val, $m)) {
				$address = $m[1];
				$name    = '';
			}
			// special case (#1489092)
			else if (preg_match('/(\s*<MAILER-DAEMON>)$/', $val, $m)) {
				$address = 'MAILER-DAEMON';
				$name    = substr($val, 0, -strlen($m[1]));
			}
			else if (preg_match('/('.$email_rx.')/', $val, $m)) {
				$name = $m[1];
			}
			else {
				$name = $val;
			}
			// dequote and/or decode name
			if ($name) {
				if ($name[0] == '"' && $name[strlen($name)-1] == '"') {
					$name = substr($name, 1, -1);
					$name = stripslashes($name);
				}
				if ($decode) {
					$name = self::decode_header($name, $fallback);
					// some clients encode addressee name with quotes around it
					if ($name[0] == '"' && $name[strlen($name)-1] == '"') {
						$name = substr($name, 1, -1);
					}
				}
			}
			if (!$address && $name) {
				$address = $name;
				$name    = '';
			}
			if ($address) {
				$address      = self::fix_email($address);
				$result[$key] = array('name' => $name, 'address' => $address);
			}
		}
		return $result;
	}

	/**
	 * Explodes header (e.g. address-list) string into array of strings
	 * using specified separator characters with proper handling
	 * of quoted-strings and comments (RFC2822)
	 *
	 * @param string $separator       String containing separator characters
	 * @param string $str             Header string
	 * @param bool   $remove_comments Enable to remove comments
	 *
	 * @return array Header items
	 */
	public static function explode_header_string($separator, $str, $remove_comments = false)
	{
		$length  = strlen($str);
		$result  = array();
		$quoted  = false;
		$comment = 0;
		$out     = '';
		for ($i=0; $i<$length; $i++) {
			// we're inside a quoted string
			if ($quoted) {
				if ($str[$i] == '"') {
					$quoted = false;
				}
				else if ($str[$i] == "\\") {
					if ($comment <= 0) {
						$out .= "\\";
					}
					$i++;
				}
			}
			// we are inside a comment string
			else if ($comment > 0) {
				if ($str[$i] == ')') {
					$comment--;
				}
				else if ($str[$i] == '(') {
					$comment++;
				}
				else if ($str[$i] == "\\") {
					$i++;
				}
				continue;
			}
			// separator, add to result array
			else if (strpos($separator, $str[$i]) !== false) {
				if ($out) {
					$result[] = $out;
				}
				$out = '';
				continue;
			}
			// start of quoted string
			else if ($str[$i] == '"') {
				$quoted = true;
			}
			// start of comment
			else if ($remove_comments && $str[$i] == '(') {
				$comment++;
			}
			if ($comment <= 0) {
				$out .= $str[$i];
			}
		}
		if ($out && $comment <= 0) {
			$result[] = $out;
		}
		return $result;
	}

	/**
	 * Interpret a format=flowed message body according to RFC 2646
	 *
	 * @param string $text Raw body formatted as flowed text
	 * @param string $mark Mark each flowed line with specified character
	 *
	 * @return string Interpreted text with unwrapped lines and stuffed space removed
	 */
	public static function unfold_flowed($text, $mark = null)
	{
		$text    = preg_split('/\r?\n/', $text);
		$last    = -1;
		$q_level = 0;
		$marks   = array();
		foreach ($text as $idx => $line) {
			if ($q = strspn($line, '>')) {
				// remove quote chars
				$line = substr($line, $q);
				// remove (optional) space-staffing
				if ($line[0] === ' ') $line = substr($line, 1);
				// The same paragraph (We join current line with the previous one) when:
				// - the same level of quoting
				// - previous line was flowed
				// - previous line contains more than only one single space (and quote char(s))
				if ($q == $q_level
						&& isset($text[$last]) && $text[$last][strlen($text[$last])-1] == ' '
						&& !preg_match('/^>+ {0,1}$/', $text[$last])
						) {
							$text[$last] .= $line;
							unset($text[$idx]);
							if ($mark) {
								$marks[$last] = true;
							}
						}
						else {
							$last = $idx;
						}
			}
			else {
				if ($line == '-- ') {
					$last = $idx;
				}
				else {
					// remove space-stuffing
					if ($line[0] === ' ') $line = substr($line, 1);
					if (isset($text[$last]) && $line && !$q_level
							&& $text[$last] != '-- '
							&& $text[$last][strlen($text[$last])-1] == ' '
							) {
								$text[$last] .= $line;
								unset($text[$idx]);
								if ($mark) {
									$marks[$last] = true;
								}
							}
							else {
								$text[$idx] = $line;
								$last = $idx;
							}
				}
			}
			$q_level = $q;
		}
		if (!empty($marks)) {
			foreach (array_keys($marks) as $mk) {
				$text[$mk] = $mark . $text[$mk];
			}
		}
		return implode("\r\n", $text);
	}

	/**
	 * Wrap the given text to comply with RFC 2646
	 *
	 * @param string $text    Text to wrap
	 * @param int    $length  Length
	 * @param string $charset Character encoding of $text
	 *
	 * @return string Wrapped text
	 */
	public static function format_flowed($text, $length = 72, $charset=null)
	{
		$text = preg_split('/\r?\n/', $text);
		foreach ($text as $idx => $line) {
			if ($line != '-- ') {
				if ($level = strspn($line, '>')) {
					// remove quote chars
					$line = substr($line, $level);
					// remove (optional) space-staffing and spaces before the line end
					$line = rtrim($line, ' ');
					if ($line[0] === ' ') $line = substr($line, 1);
					$prefix = str_repeat('>', $level) . ' ';
					$line   = $prefix . self::wordwrap($line, $length - $level - 2, " \r\n$prefix", false, $charset);
				}
				else if ($line) {
					$line = self::wordwrap(rtrim($line), $length - 2, " \r\n", false, $charset);
					// space-stuffing
					$line = preg_replace('/(^|\r\n)(From| |>)/', '\\1 \\2', $line);
				}
				$text[$idx] = $line;
			}
		}
		return implode("\r\n", $text);
	}

	/**
	 * Improved wordwrap function with multibyte support.
	 * The code is based on Zend_Text_MultiByte::wordWrap().
	 *
	 * @param string $string      Text to wrap
	 * @param int    $width       Line width
	 * @param string $break       Line separator
	 * @param bool   $cut         Enable to cut word
	 * @param string $charset     Charset of $string
	 * @param bool   $wrap_quoted When enabled quoted lines will not be wrapped
	 *
	 * @return string Text
	 */
	public static function wordwrap($string, $width=75, $break="\n", $cut=false, $charset=null, $wrap_quoted=true)
	{
		// Note: Never try to use iconv instead of mbstring functions here
		//       Iconv's substr/strlen are 100x slower (#1489113)
		if ($charset && $charset != WSC_CHARSET) {
			mb_internal_encoding($charset);
		}
		// Convert \r\n to \n, this is our line-separator
		$string       = str_replace("\r\n", "\n", $string);
		$separator    = "\n"; // must be 1 character length
		$result       = array();
		while (($stringLength = mb_strlen($string)) > 0) {
			$breakPos = mb_strpos($string, $separator, 0);
			// quoted line (do not wrap)
			if ($wrap_quoted && $string[0] == '>') {
				if ($breakPos === $stringLength - 1 || $breakPos === false) {
					$subString = $string;
					$cutLength = null;
				}
				else {
					$subString = mb_substr($string, 0, $breakPos);
					$cutLength = $breakPos + 1;
				}
			}
			// next line found and current line is shorter than the limit
			else if ($breakPos !== false && $breakPos < $width) {
				if ($breakPos === $stringLength - 1) {
					$subString = $string;
					$cutLength = null;
				}
				else {
					$subString = mb_substr($string, 0, $breakPos);
					$cutLength = $breakPos + 1;
				}
			}
			else {
				$subString = mb_substr($string, 0, $width);
				// last line
				if ($breakPos === false && $subString === $string) {
					$cutLength = null;
				}
				else {
					$nextChar = mb_substr($string, $width, 1);
					if ($nextChar === ' ' || $nextChar === $separator) {
						$afterNextChar = mb_substr($string, $width + 1, 1);
						// Note: mb_substr() does never return False
						if ($afterNextChar === false || $afterNextChar === '') {
							$subString .= $nextChar;
						}
						$cutLength = mb_strlen($subString) + 1;
					}
					else {
						$spacePos = mb_strrpos($subString, ' ', 0);
						if ($spacePos !== false) {
							$subString = mb_substr($subString, 0, $spacePos);
							$cutLength = $spacePos + 1;
						}
						else if ($cut === false) {
							$spacePos = mb_strpos($string, ' ', 0);
							if ($spacePos !== false && ($breakPos === false || $spacePos < $breakPos)) {
								$subString = mb_substr($string, 0, $spacePos);
								$cutLength = $spacePos + 1;
							}
							else if ($breakPos === false) {
								$subString = $string;
								$cutLength = null;
							}
							else {
								$subString = mb_substr($string, 0, $breakPos);
								$cutLength = $breakPos + 1;
							}
						}
						else {
							$cutLength = $width;
						}
					}
				}
			}
			$result[] = $subString;
			if ($cutLength !== null) {
				$string = mb_substr($string, $cutLength, ($stringLength - $cutLength));
			}
			else {
				break;
			}
		}
		if ($charset && $charset != WSC_CHARSET) {
			mb_internal_encoding(WSC_CHARSET);
		}
		return implode($break, $result);
	}

	/**
	 * A method to guess the mime_type of an attachment.
	 *
	 * @param string  $path        Path to the file or file contents
	 * @param string  $name        File name (with suffix)
	 * @param string  $failover    Mime type supplied for failover
	 * @param boolean $is_stream   Set to True if $path contains file contents
	 * @param boolean $skip_suffix Set to True if the config/mimetypes.php mappig should be ignored
	 *
	 * @return string
	 * @author Till Klampaeckel <till@php.net>
	 * @see    http://de2.php.net/manual/en/ref.fileinfo.php
	 * @see    http://de2.php.net/mime_content_type
	 */
	public static function file_content_type($path, $name, $failover = 'application/octet-stream', $is_stream = false, $skip_suffix = false)
	{
		$tipoMime = $failover;

		$realpath = realpath($path);
		if (
			$realpath
			&& function_exists( 'finfo_file' )
			&& function_exists( 'finfo_open' )
			&& defined( 'FILEINFO_MIME_TYPE' )
		)
		{
			// Usa  la extensin Fileinfo PECL (PHP >=5.3)
			$tipoMime = finfo_file( finfo_open( FILEINFO_MIME_TYPE ), $realpath );
		}
		else //Lo calculamos por la extensin
		{
			$fileExtension = pathinfo($path, PATHINFO_EXTENSION);

			//Vector de tipos MIME
			$mime_types = array
			(
					//texto
					'txt' => 'text/plain',
					'htm' => 'text/html',
					'html' => 'text/html',
					'php' => 'text/html',
					'css' => 'text/css',
					'js' => 'application/javascript',
					'json' => 'application/json',
					'xml' => 'application/xml',
					'csv' => 'text/csv',

					// imgenes
					'png' => 'image/png',
					'jpe' => 'image/jpeg',
					'jpeg' => 'image/jpeg',
					'jpg' => 'image/jpeg',
					'gif' => 'image/gif',
					'bmp' => 'image/bmp',
					'ico' => 'image/vnd.microsoft.icon',
					'tiff' => 'image/tiff',
					'tif' => 'image/tiff',
					'svg' => 'image/svg+xml',
					'svgz' => 'image/svg+xml',

					// archivos
					'zip' => 'application/zip',
					'rar' => 'application/x-rar-compressed',
					'exe' => 'application/x-msdownload',
					'msi' => 'application/x-msdownload',
					'cab' => 'application/vnd.ms-cab-compressed',
					'tgz' => 'application/tar+gzip',
					'tar.gz' => 'application/tar+gzip',
					'tar' => 'application/tar',
					'gz' => 'application/gzip',
					'7z' => 'application/x-7z-compressed',
					's7z' => 'application/x-7z-compressed',

					// audio/video
					'mp3' => 'audio/mpeg',
					'qt' => 'video/quicktime',
					'mov' => 'video/quicktime',
					'mpeg' => 'video/mpeg',
					'avi' => 'video/x-msvideo',
					'swf' => 'application/x-shockwave-flash',
					'flv' => 'video/x-flv',

					// adobe
					'pdf' => 'application/pdf',
					'psd' => 'image/vnd.adobe.photoshop',
					'ai' => 'application/postscript',
					'eps' => 'application/postscript',
					'ps' => 'application/postscript',

					// MSOffice
					'doc' => 'application/msword',
					'dot' => 'application/msword',
					'docx' => 'application/msword',
					'rtf' => 'application/rtf',

					'xls' => 'application/vnd.ms-excel',
					'xlsx' => 'application/vnd.ms-excel',
					'xlm' => 'application/vnd.ms-excel',
					'xla' => 'application/vnd.ms-excel',
					'xlc' => 'application/vnd.ms-excel',
					'xlt' => 'application/vnd.ms-excel',
					'xlw' => 'application/vnd.ms-excel',

					'ppt' => 'application/vnd.ms-powerpoint',
					'pptx' => 'application/vnd.ms-powerpoint',
					'pps' => 'application/vnd.ms-powerpoint',
					'pot' => 'application/vnd.ms-powerpoint',

					// libreOffice
					'odc' => 'application/vnd.oasis.opendocument.chart',
					'otc' => 'application/vnd.oasis.opendocument.chart-template',
					'odf' => 'application/vnd.oasis.opendocument.formula',
					'otf' => 'application/vnd.oasis.opendocument.formula-template',
					'odg' => 'application/vnd.oasis.opendocument.graphics',
					'otg' => 'application/vnd.oasis.opendocument.graphics-template',
					'odi' => 'application/vnd.oasis.opendocument.image',
					'oti' => 'application/vnd.oasis.opendocument.image-template',
					'odp' => 'application/vnd.oasis.opendocument.presentation',
					'otp' => 'application/vnd.oasis.opendocument.presentation-template',
					'ods' => 'application/vnd.oasis.opendocument.spreadsheet',
					'ots' => 'application/vnd.oasis.opendocument.spreadsheet-template',
					'odt' => 'application/vnd.oasis.opendocument.text',
					'otm' => 'application/vnd.oasis.opendocument.text-master',
					'ott' => 'application/vnd.oasis.opendocument.text-template',
					'oth' => 'application/vnd.oasis.opendocument.text-web',

					//VCards...
					'vcf' => 'text/vcard',
					'ics' => 'text/calendar',
			);

			//Si la extensin se contempla en nuestro vector de tipos...
			if (array_key_exists($fileExtension, $mime_types))
			{
				$tipoMime = $mime_types[$fileExtension];
			}
		}
		return $tipoMime;
	}//Fin funcion



	/**
	 * Detect image type of the given binary data by checking magic numbers.
	 *
	 * @param string $data  Binary file content
	 *
	 * @return string Detected mime-type or jpeg as fallback
	 */
	public static function image_content_type($data)
	{
		$type = 'jpeg';
		if      (preg_match('/^\x89\x50\x4E\x47/', $data)) $type = 'png';
		else if (preg_match('/^\x47\x49\x46\x38/', $data)) $type = 'gif';
		else if (preg_match('/^\x00\x00\x01\x00/', $data)) $type = 'ico';
		//  else if (preg_match('/^\xFF\xD8\xFF\xE0/', $data)) $type = 'jpeg';
		return 'image/' . $type;
	}


	/**
	 * Try to fix invalid email addresses
	 */
	public static function fix_email($email)
	{
		$m = array();
		$parts = array();
		$delimiter = '@';
		$strlen = strlen($email);
		for ($q=$p=$i=0; $i < $strlen; $i++)
		{
			if ($email[$i] == "\"" && $email[$i-1] != "\\")
			{
				$q = $q ? false : true;
			}
			else if (!$q && preg_match("/$delimiter/", $email[$i]))
			{
				$parts[] = substr($email, $p, $i - $p);
				$p = $i + 1;
			}
		}
		$parts[] = (string) substr($email, $p);

		foreach ($parts as $idx => $part)
		{
			// remove redundant quoting (#1490040)
			if ($part[0] == '"' && preg_match('/^"([a-zA-Z0-9._+=-]+)"$/', $part, $m)) {
				$parts[$idx] = $m[1];
			}
		}
		return implode('@', $parts);
	}//fin fix_email
}